# adapter.py

# Copyright 2022-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import sys
import ssl
import httpx
import socket
import random
from httpx import Response
from typing import Any, List, Optional
from prometheus_client.parser import text_string_to_metric_families as metric_parser
from sseapiclient import APIClient
import aria.ops.adapter_logging as logging
from aria.ops.object import Object
from aria.ops.adapter_instance import AdapterInstance
from aria.ops.definition.adapter_definition import AdapterDefinition
from aria.ops.result import CollectResult, EndpointResult, TestResult, RelationshipUpdateModes
from aria.ops.suite_api_client import key_to_object
from aria.ops.suite_api_client import SuiteApiClient
from aria.ops.timer import Timer
import constants
from constants import ADAPTER_KIND, ADAPTER_NAME

logger = logging.getLogger(__name__)


def _get_validated_certificate_for_host(hostname: str, valid_certs: List[dict]) -> Optional[str]:
    if not valid_certs:
        logger.warning(f"No valid certificates are present in the adapter instance. Cannot validate certificate for {hostname}.")
        return None
    logger.debug(f"List of accepted certificate specs in adapter:\n{valid_certs}")
    if ":" in hostname:
        host, port_str = hostname.split(":", maxsplit=1)
        try:
            port = int(port_str)
        except ValueError as e:
            port = 443
    else:
        host = hostname
        port = 443
    try:
        host_cert = str(ssl.get_server_certificate((host, port)))
        logger.debug(f"Certificate presented by {hostname}:\n{host_cert}")
    except ssl.SSLError as e:
        logger.warning(f"Could not retrieve certificate for {hostname}: {e}")
        return None
    for cert in valid_certs:
        if cert.get('cert_pem_string','').replace("\n", "").strip() == host_cert.replace("\n", "").strip():
            #.replace("\n", "") above is required as a workaround to valid_certs not containing newlines in the cert body for some reason when executing in a real Aria Ops instance.
            logger.debug(f"Certificate presented by {hostname} was found in the following spec:\n{cert}")
            return host_cert
    logger.warning(f"Certificate presented by {hostname} does not match any of the accepted certificates in the adapter instance.")
    return None


def _get_metrics(adapter_instance: AdapterInstance) -> List[Response]:
    logger.debug(f"Preparing to collect metrics from Adapter instance: {adapter_instance.get_json()}")
    hosts = adapter_instance.get_identifier_value(constants.HOST_IDENTIFIER)
    username = adapter_instance.get_credential_value(constants.METRIC_USER_CREDENTIAL)
    password = adapter_instance.get_credential_value(constants.METRIC_PASSWORD_CREDENTIAL)
    #certificate = _get_validated_certificate_for_host(host, adapter_instance.certificates)
    # Check that we have values for all the identifiers/credentials we need
    none_check = []
    if hosts is None:
        none_check.append(f"No host found with identifier \"{constants.HOST_IDENTIFIER}\"")
    else:
        hostsList = [host.strip() for host in hosts.split(',') if host.strip()]
    if username is None:
        none_check.append(f"No username found with identifier \"{constants.METRIC_USER_CREDENTIAL}\"")
    if password is None:
        none_check.append(f"No password found with identifier \"{constants.METRIC_PASSWORD_CREDENTIAL}\"")
    for host in hostsList:
        certificate = _get_validated_certificate_for_host(host, adapter_instance.certificates)
        if certificate is None:
            none_check.append(f"Failed to validate certificate for host \"{host}\". See logs for details.")
    if none_check:
        error_msg = f"Failed to retrieve the following required properties in adapter \"{ADAPTER_NAME}\": {', '.join(none_check)}"
        logger.error(error_msg)
        raise Exception(error_msg)
    
    responses = []
    for host in hostsList:
        metrics_url = f"{constants.HOST_PROTOCOL}://{host}/{constants.HOST_URL}"
        logger.debug(f"Retrieving metrics using URL: {metrics_url} User: {username}")

        # Fetch the metrics
        try:
            #Verify set to False to allow self-signed certificates or certificates with a hostname that doesn't match the hostname used to connect to the target.
            #Additionally, certificate verification is done manually above by comparing the certificate presented by the target to the list of accepted certificates in the adapter instance.
            with httpx.Client(auth=(username, password), verify=False) as client:
                response = client.get(metrics_url, timeout=60.0)
            logger.debug(f"Request status code: {response.status_code}")
            if response.status_code == 200:
                logger.debug("Request response line count: {}".format(response.text.count('\n')))
            responses.append(response)
        except httpx.RequestError as e:
            logger.error(f"Request error occurred during metrics collection: {e}")
            raise
        except httpx.HTTPStatusError as e:
            logger.error(f"HTTP error occurred during metrics collection: {e}")
            raise
        except Exception as e:
            logger.error(f"Unexpected error occurred during metrics collection: {e}")
            raise
    return responses


def _get_master_cluster_map(adapter_instance: AdapterInstance) -> dict:
    logger.debug(f"Preparing to collect metrics from Adapter instance: {adapter_instance.get_json()}")
    hosts = adapter_instance.get_identifier_value(constants.HOST_IDENTIFIER)
    username = adapter_instance.get_credential_value(constants.EAPI_USER_CREDENTIAL)
    password = adapter_instance.get_credential_value(constants.EAPI_PASSWORD_CREDENTIAL)
    
    none_check = []
    if hosts is None:
        none_check.append(f"No host found with identifier \"{constants.HOST_IDENTIFIER}\"")
    else:
        hostsList = [host.strip() for host in hosts.split(',') if host.strip()]
    if username is None:
        none_check.append(f"No username found with identifier \"{constants.EAPI_USER_CREDENTIAL}\"")
    if password is None:
        none_check.append(f"No password found with identifier \"{constants.EAPI_PASSWORD_CREDENTIAL}\"")
    if none_check:
        error_msg = f"Failed to retrieve the following required properties in adapter \"{ADAPTER_NAME}\": {', '.join(none_check)}"
        logger.error(error_msg)
        raise Exception(error_msg)
    
    raasHost = random.choice(hostsList)
    # Fetch the metrics
    try:
        client = APIClient(f"https://{raasHost}", username, password, config_name="internal", timeout=120, ssl_validate_cert=False)
        ret = client.api.master.get_master_presence().ret
        master_cluster_mapping = {result['master_id']: result['cluster_id'] for result in ret['results']}
    except Exception as e:
        logger.error(f"Unexpected error occurred while connecting to and/or querying '{raasHost}': {e}")
        raise
    return master_cluster_mapping


def _process_sample(sample_object, family):
    if family.type == "histogram":
        if sample_object.name.endswith("_sum") or sample_object.name.endswith("_count"):
            metric_name = sample_object.name
        else:
            return None, None
    else:
        metric_name = family.name
    labels = sample_object.labels
    if "task" in labels:
        metric_name = f"{metric_name}|{labels['task']}"
    elif "method" in labels:
        if "resource" in labels:
            metric_name = f"{metric_name}|{labels['resource']}|{labels['method']}"
        else:
            metric_name = f"{metric_name}|{labels['method']}"
    return metric_name, sample_object.value


def _get_hostname(resource_name: str) -> str:
    try:
        # Check if the resource name is an IP address
        socket.inet_aton(resource_name)
        try:
            # Try to resolve the IP address to a hostname
            fqdn = socket.gethostbyaddr(resource_name)[0]
            # Extract the hostname from the FQDN
            hostname = fqdn.split('.')[0]
            return hostname
        except socket.herror:
            # IP address cannot be resolved
            return None
    except socket.error:
        # Not an IP, check for FQDN and extract the hostname
        if '.' in resource_name:
            return resource_name.split('.')[0]
        else:
            # Already a hostname, return as is
            return resource_name


def _get_vm_object(
    adapter_instance: AdapterInstance,
    resourceName: str #Note, this could be a Hostname, FQDN, or IP address
) -> Optional[Object]:

    if '.' in resourceName:
        vmName = _get_hostname(resourceName)
    else:
        vmName = resourceName

    if not vmName:
        logger.debug(f"Could not resolve resource name '{resourceName}' to a valid VM name.")
        return None
    
    with adapter_instance.suite_api_client as client:
        searchPattern = f"(?i)^{vmName}$"
        vms: List[Object] = client.query_for_resources(
            {
                "adapterKind": [constants.VCENTER_ADAPTER_KIND],
                "resourceKind": ["VirtualMachine"],
                "regex": [searchPattern]
            }
        )

    if not vms or len(vms) != 1:
        vms_count = 0 if vms is None else len(vms)
        logger.warn(f"{vms_count} VMs found across all '{constants.VCENTER_ADAPTER_KIND}' adapters with the searchPattern '{searchPattern}'. Exactly 1 match is required.")
        vm = None
    else:
        vm = vms[0]
        logger.debug(f"VM found with the name '{vm.get_key().name}'.")
        logger.debug(f"VM JSON: {vm.get_json()}")
        logger.debug(f"VM Children: {vm.get_children()}")
        logger.debug(f"VM Key JSON: {vm.get_key().get_json()}")

    return vm


def _get_set_cluster_property(
    adapter_instance: AdapterInstance,
    vmObj: Object
) -> Optional[str]:
    with adapter_instance.suite_api_client as client:
        resourceID = _get_object_id(client, vmObj)
        clusterName = _get_property_value(client, vmObj, constants.SALT_CLUSTER_PROP_KEY, resourceID)
        if clusterName is None:
            propertyList = adapter_instance.get_property(constants.MASTER_CLUSTER_MAP_KEY)
            if propertyList is None or len(propertyList) < 1:
                clusterMap = _get_master_cluster_map(adapter_instance)
                logger.debug(f"Successfully retrieved cluster map from eAPI. Adding it to this adapter instance. Map:\n{clusterMap}")
                adapter_instance.with_property(constants.MASTER_CLUSTER_MAP_KEY, clusterMap)
            elif len(propertyList) > 1:
                logger.warn(f"{len(propertyList)} Properties found with key {constants.MASTER_CLUSTER_MAP_KEY} in our adapter_instance. Returning None for clsuterName to be safe.")
                return None
            else:
                clusterMap = propertyList[0].value
                logger.debug(f"Found cluster map in this adapter instance. Map:\n{clusterMap}")
            #Select Cluster Name from map
            matching_keys = [key for key in clusterMap.keys() if vmObj.get_key().name in key]
            if len(matching_keys) != 1:
                logger.warn(f"{len(matching_keys)} matches found for Master with name '{vmObj.get_key().name}' in clusterMap. Exactly 1 required. Cannot determine the correct cluster.")
                return None
            else:
                clusterName = clusterMap[matching_keys[0]]
            if clusterName is not None:
                vmObj.with_property(constants.SALT_CLUSTER_PROP_KEY, clusterName)
        elif clusterName == "":
            logger.warn("There was an issue retrieving the Cluster ID from the VM object using the suite-api. See logs above.")
            clusterName = None
    return clusterName


def _preserve_existing_object_relationships(
    adapter_instance: AdapterInstance,
    vmObj: Object
) -> List[Object]:

    vcid = vmObj.get_identifier_value(constants.VM_VCENTER_IDENTIFIER)

    if not vcid:
        logger.debug(f"Could not find a '{constants.VM_VCENTER_IDENTIFIER}' indentifier on VM object with name '{vmObj.get_key().name}'. Cannot check existing relationships.")
        return None
    
    with adapter_instance.suite_api_client as client:
        vcAdapterId = _get_vcenter_adapter_instance_id(client, vcid)
        vmObjId = _get_object_id(client, vmObj, vcAdapterId)
        children: List[Object] = client.query_for_resources(
            {
                "parentId": [vmObjId]
            }
        )

    if children is None or len(children) < 1:
        logger.debug(f"No children found for VM object with name '{vmObj.get_key().name}'. No relationships to preserve.")
        return None
    else:
        for child in children:
            logger.debug(f"Adding Child object with name '{child.get_key().name}' to VM object '{vmObj.get_key().name}'.")
            vmObj.add_child(child)

    return children


def _get_vcenter_adapter_instance_id(
    client: SuiteApiClient, vcid: str
) -> Optional[str]:
    ais: List[Object] = client.query_for_resources(
        {
            "adapterKind": [constants.VCENTER_ADAPTER_KIND],
            "resourceKind": ["VMwareAdapter Instance"],
        }
    )
    for ai in ais:
        aiVcid = ai.get_identifier_value(constants.VM_VCENTER_IDENTIFIER)
        logger.debug(
            f"Considering vCenter Adapter Instance with '{constants.VM_VCENTER_IDENTIFIER}': {aiVcid}"
        )
        if aiVcid == vcid:
            return _get_adapter_instance_id(client, ai)
    return None


def _get_adapter_instance_id(
    client: SuiteApiClient, adapter_instance: Object
) -> Optional[Any]:
    response = client.get(
        f"api/adapters?adapterKindKey={adapter_instance.get_key().adapter_kind}"
    )
    if response.status_code < 300:
        for ai in json.loads(response.content).get("adapterInstancesInfoDto", []):
            adapter_instance_key = key_to_object(ai.get("resourceKey")).get_key()
            if adapter_instance_key == adapter_instance.get_key():
                return ai.get("id")
    return None


def _get_property_value(
    client: SuiteApiClient, obj: Object, propKey: str, resourceID: Optional[str]
) -> Optional[str]:
    #TODO:Needs optimization: resourceID of VM objects have already been queried elsewhere. Need to work out a way to persist this ID so we dont have to retrieve it more than once.
    if resourceID is None:
        resourceID = _get_object_id(client, obj)
    query = {
        "resourceIds" : [ resourceID ],
        "propertyKeys" : [ propKey ]
    }
    response = client.post(
        url=f"api/resources/properties/latest/query",
        json=query
    )
    propVal = None
    if response.status_code < 300:
        properties = json.loads(response.content).get("values", [])
        if len(properties) != 1:
            logger.debug(f"{len(properties)} Properties found matching key '{propKey}' on VM '{obj.get_key().name}'. Exactly 1 match is required.")
            return None if len(properties) < 1 else ""
        try:
            prop_content = properties[0].get("property-contents", {}).get("property-content", [])
            if prop_content and isinstance(prop_content, list):
                values = prop_content[0].get("values", [])
                if values and isinstance(values, list):
                    propVal = values[0]
                else:
                    propVal = ""
            else:
                propVal = ""
        except (IndexError, TypeError) as e:
            logger.exception(f"Failed retrieving the property value from response body: {e}")
            propVal = ""
    else:
        logger.warn(f"Property query failed with response code '{response.status_code}'. "
                    f"Response content: '{response.text}', ")
    return propVal


def _get_object_id(
    client: SuiteApiClient, resObj: Object, adapterId: Optional[str] = None
) -> Optional[Any]:
    if adapterId is None:
        vcid = resObj.get_identifier_value(constants.VM_VCENTER_IDENTIFIER)
        if not vcid:
            logger.debug(f"Could not find a '{constants.VM_VCENTER_IDENTIFIER}' indentifier on object with name '{resObj.get_key().name}'. Cannot get object id as a result.")
            return None
        adapterId = _get_vcenter_adapter_instance_id(client, vcid)
    response = client.get(
        f"api/resources?adapterInstanceId={adapterId}&regex=^{resObj.get_key().name}$&resourceKind={resObj.object_type()}"
    )
    if response.status_code < 300:
        for res in json.loads(response.content).get("resourceList", []):
            res_obj_key = key_to_object(res.get("resourceKey")).get_key()
            if res_obj_key == resObj.get_key():
                return res.get("identifier")
    return None


def _result_add_unique_objects(
    result: CollectResult, objects: List[Object]
) -> None:
    for obj in objects:
        if result.get_object(obj.get_key()) is None:
            result.add_object(obj)


# Function to get adapter definition
def get_adapter_definition() -> AdapterDefinition:
    """
    The adapter definition defines the object types and attribute types (metric/property) that are present
    in a collection. Setting these object types and attribute types helps VMware Aria Operations to
    validate, process, and display the data correctly.
    :return: AdapterDefinition
    """
    with Timer(logger, "Get Adapter Definition"):
        definition = AdapterDefinition(ADAPTER_KIND, ADAPTER_NAME)

        definition.define_string_parameter(
            constants.HOST_IDENTIFIER,
            label="RaaS Hostname(s)",
            description=(
                "<p><strong>Hostnames/FQDNs/IPs:</strong>Comma-separated list of hostnames, "
                "FQDNs, or IP addresses for Aria Automation Config (RaaS) Servers.</p>"
                "<p>If a server uses a port other than the default <strong>443</strong>, include the port number "
                "in the format <code>'hostname:port'</code>.<br>Example: "
                "<code>'server1.example.com, 192.168.1.2:8080, server2'</code>.</p>"
            ),
            required=True
        )

        metric_credential = definition.define_credential_type("metric_cred", "Metric Credential")
        metric_credential.define_string_parameter(constants.METRIC_USER_CREDENTIAL, "Metric Username", True)
        metric_credential.define_password_parameter(constants.METRIC_PASSWORD_CREDENTIAL, "Metric Password", True)
        config_credential = definition.define_credential_type("metric_plus_config_creds", "Metric & eAPI Credentials")
        config_credential.define_string_parameter(constants.METRIC_USER_CREDENTIAL, "Metric Username", True)
        config_credential.define_password_parameter(constants.METRIC_PASSWORD_CREDENTIAL, "Metric Password", True)
        config_credential.define_string_parameter(constants.EAPI_USER_CREDENTIAL, "eAPI (RaaS) Username", True)
        config_credential.define_password_parameter(constants.EAPI_PASSWORD_CREDENTIAL, "eAPI (RaaS) Password", True)
        
        # definition.define_enum_parameter(
        #     constants.PROP_DISCOVERY_IDENTIFIER,
        #     label="Advanced OS Property Discovery",
        #     description="Requires Telegraf agent to be installed/configured on each VM related to this management pack under "
        #                 "Environment -> Applications -> Manage Telegraf Agents. Uses the agent to query the Guest OS for details not otherwise exposed to Aria Operations.",
        #     required=True,
        #     advanced=True,
        #     values=["enabled","disabled"],
        #     default="enabled"
        # )

        # The key 'container_memory_limit' is a special key read by the VMware Aria Operations
        # collector to determine how much memory to allocate to the docker container running
        # this adapter. It does not need to be read inside the adapter code. However, removing
        # the definition from the object model will remove the ability to change the container
        # memory limit during the adapter's configuration, and the VMware Aria Operations collector
        # will give 1024 MB of memory to the container running the adapter instance.
        definition.define_int_parameter(
            "container_memory_limit",
            label="Adapter Memory Limit (MB)",
            description="Sets the maximum amount of memory VMware Aria Operations can "
            "allocate to the container running this adapter instance.",
            required=True,
            advanced=True,
            default=1024,
        )        
        
        # TODO: Add object types, including identifiers, metrics, and properties
        # Define object and attribute types here
        # For example:
        # definition.define_metric('celery_tasks_queued_total', label='Celery Tasks Queued', description='Total number of queued Celery tasks', unit='count')
        config_object = definition.define_object_type(constants.CONFIG_OBJECT_KIND, constants.CONFIG_OBJECT_NAME)
        master_object = definition.define_object_type(constants.MASTER_OBJECT_KIND, constants.MASTER_OBJECT_NAME)
        config_object.define_string_property("description", "Description")
        master_object.define_string_property("description", "Description")

        logger.debug(f"Returning adapter definition: {definition.to_json()}")
        return definition


# Function to test adapter connectivity
def test(adapter_instance: AdapterInstance) -> TestResult:
    with Timer(logger, "Test"):
        result = TestResult()
        try:
            errors = []
            # Test connecting to eAPI
            if adapter_instance.get_credential_type() == constants.EAPI_CREDENTIAL_TYPE:
                #get master clusters
                cluster_map = _get_master_cluster_map(adapter_instance)
                if cluster_map is None:
                    errors.append(f"Failed querying the Aria Config eAPI with the specified hostname(s) and/or credentials.")
            
            # Test retrieving metrics
            metricResponses = _get_metrics(adapter_instance)

            for response in metricResponses:
                # Check if the response is successful
                if response.status_code != 200:
                    errors.append(f"Connection to '{response.url}' failed with status code: {response.status_code}")
            if errors:
                # Atleast one connection failed, set the TestResult to error
                result.with_error("\n".join(errors))
            else:
                result.is_success()
        except Exception as e:
            logger.error("Unexpected connection test error")
            logger.exception(e)
            result.with_error("Unexpected connection test error: " + repr(e))
        finally:
            logger.debug(f"Returning test result: {result.get_json()}")
            return result


# Function to collect metrics data
def collect(adapter_instance: AdapterInstance) -> CollectResult:
    with Timer(logger, "Collection"):
        result = CollectResult( target_definition = get_adapter_definition() )
        result.update_relationships = RelationshipUpdateModes.PER_OBJECT
        try:
            # Connect to the metrics endpoint and retrieve the data
            metricResponses = _get_metrics(adapter_instance)
            [response.raise_for_status() for response in metricResponses] # Will raise an HTTPError if the HTTP request returned an unsuccessful status code

            #Detect if the adapter has Advanced property discovery enabled.
            credType = adapter_instance.get_credential_type()
            advPropsEnabled = credType == constants.EAPI_CREDENTIAL_TYPE
            logger.debug(f"Advanced Property detection functionality enabled: {advPropsEnabled}")

            # Parse the Prometheus metrics and group them into RaaS & Salt Master objects
            global_obj = None
            raas_objects = {}
            master_objects = {}
            cluster_objects = {}
            raas_vms = []
            master_vms = []
            for metricResponse in metricResponses:
                for metric_family in metric_parser(metricResponse.text):
                    for sample in metric_family.samples:
                        labels = sample.labels
                        # Check and add 'raas_instance' if it exists in labels
                        if 'raas_instance' in labels:
                            if labels['raas_instance'] in raas_objects.keys():
                                obj = raas_objects[labels['raas_instance']]
                            else:
                                logger.debug(f"Adding additional '{constants.CONFIG_OBJECT_NAME}' object with key '{labels['raas_instance']}'.")
                                obj = result.object(ADAPTER_KIND, constants.CONFIG_OBJECT_KIND, labels['raas_instance'])
                                obj.with_property("description", "RaaS_Instance_Metrics")
                                vmObj = _get_vm_object(adapter_instance, labels['raas_instance'])
                                if vmObj:
                                    logger.debug(f"Starting relationship preservation for '{vmObj.get_key().name}'.")
                                    vmChildren = _preserve_existing_object_relationships(adapter_instance, vmObj)
                                    _result_add_unique_objects(result, vmChildren)
                                    logger.debug(f"Adding VM object relationship for '{labels['raas_instance']}'.")
                                    vmObj.add_child(obj)
                                    vmObj = result.add_object(vmObj)
                                    raas_vms.append(vmObj)
                                logger.debug(f"New Object JSON: {obj.get_json()}")
                                raas_objects[labels['raas_instance']] = obj
                        # Check and add 'master_id' if it exists in labels
                        elif 'master_id' in labels:
                            if labels['master_id'] in master_objects.keys():
                                obj = master_objects[labels['master_id']]
                            elif labels['master_id'] in cluster_objects.keys():
                                obj = cluster_objects[labels['master_id']]
                            else:
                                logger.debug(f"Adding additional '{constants.MASTER_OBJECT_NAME}' object with key '{labels['master_id']}'.")
                                obj = result.object(ADAPTER_KIND, constants.MASTER_OBJECT_KIND, labels['master_id'])
                                # Come back and take another look at this
                                if "cluster" in labels['master_id'].lower():
                                    obj.with_property("description", "Salt_Master_Cluster_Metrics")
                                    cluster_objects[labels['master_id']] = obj
                                else:
                                    obj.with_property("description", "Salt_Master_Instance_Metrics")
                                    vmObj = _get_vm_object(adapter_instance, labels['master_id'])
                                    if vmObj:
                                        logger.debug(f"Starting relationship preservation for '{vmObj.get_key().name}'.")
                                        vmChildren = _preserve_existing_object_relationships(adapter_instance, vmObj)
                                        _result_add_unique_objects(result, vmChildren)
                                        logger.debug(f"Adding VM object relationship for '{labels['master_id']}'.")
                                        vmObj.add_child(obj)
                                        vmObj = result.add_object(vmObj)
                                        master_vms.append(vmObj)
                                    master_objects[labels['master_id']] = obj
                                logger.debug(f"New Object JSON: {obj.get_json()}")
                        else:
                            if global_obj:
                                obj = global_obj
                            else:
                                adapterKey = adapter_instance.get_key()
                                adapterName = "Global-" + adapterKey.name
                                logger.debug(f"Adding additional '{constants.CONFIG_OBJECT_NAME}' object with key '{adapterName}'.")
                                obj = result.object(ADAPTER_KIND, constants.CONFIG_OBJECT_KIND, adapterName)
                                obj.with_property("description", "RaaS_Global_Metrics")
                                logger.debug(f"New Object JSON: {obj.get_json()}")
                                global_obj = obj
                        sample_name, sample_value = _process_sample(sample, metric_family)
                        if sample_name and sample_value is not None:
                            if not obj.get_metric(sample_name):
                                #logger.debug(f"Adding sample with name '{sample_name}' and value '{sample_value}' to Object '{obj.get_key().name}'.")
                                obj.with_metric(sample_name, sample_value)
            # Add All RaaS Instances as parents to global RaaS metrics
            if raas_objects:
                logger.debug(f"Adding RaaS Instance object relationship for RaaS Global Metrics. {list(raas_objects.values())}")
                for raas_vm in raas_vms:
                    raas_vm.add_child(global_obj)
            # Add the appropriate master VMs as parents to each of the master cluster objects.
            if cluster_objects and advPropsEnabled:
                logger.debug(f"Adding Master Cluster object relationships to appropriate VMs.")
                for master_vm in master_vms:
                    clusterId = _get_set_cluster_property(adapter_instance, master_vm)
                    if clusterId is not None:
                        clusterObj = cluster_objects.get(clusterId)
                        if clusterObj is not None:
                            logger.debug(f"Adding Master Cluster '{clusterId}' as a child to VM '{master_vm.get_key().name}'.")
                            master_vm.add_child(clusterObj)
                        else:
                            logger.debug(f"No Master Cluster found with name '{clusterId}'.")
                    else:
                        logger.debug(f"No Master Cluster ID was found on the VM with name '{master_vm.get_key().name}'.")
        except httpx.HTTPStatusError as e:
            logger.error("HTTP error occurred during collection")
            logger.exception(e)
            result.with_error(f"HTTP error occurred: {repr(e)}")
        except httpx.RequestError as e:
            logger.error("Request error occurred during collection")
            logger.exception(e)
            result.with_error(f"Request error occurred: {repr(e)}")
        except Exception as e:
            logger.error("Unexpected collection error")
            logger.exception(e)
            result.with_error("Unexpected collection error: " + repr(e))
        finally:
            logger.debug(f"Returning collection result {result.get_json()}")
            return result


# Function to get endpoints for SSL certificate handling
def get_endpoints(adapter_instance: AdapterInstance) -> EndpointResult:
    with Timer(logger, "Get Endpoints"):
        result = EndpointResult()
        # In the case that an SSL Certificate is needed to communicate to the target,
        # add each URL that the adapter uses here. Often this will be derived from a
        # 'host' parameter in the adapter instance. In this Adapter we don't use any
        # HTTPS connections, so we won't add any. If we did, we might do something like
        # this:
        # result.with_endpoint(adapter_instance.get_identifier_value("host"))
        #
        # Multiple endpoints can be returned, like this:
        # result.with_endpoint(adapter_instance.get_identifier_value("primary_host"))
        # result.with_endpoint(adapter_instance.get_identifier_value("secondary_host"))
        #
        # This 'get_endpoints' method will be run before the 'test' method,
        # and VMware Aria Operations will use the results to extract a certificate from
        # each URL. If the certificate is not trusted by the VMware Aria Operations
        # Trust Store, the user will be prompted to either accept or reject the
        # certificate. If it is accepted, the certificate will be added to the
        # AdapterInstance object that is passed to the 'test' and 'collect' methods.
        # Any certificate that is encountered in those methods should then be validated
        # against the certificate(s) in the AdapterInstance.

        # Dynamically retrieve the endpoint URL from the adapter instance
        hosts = adapter_instance.get_identifier_value(constants.HOST_IDENTIFIER)
        hostsList = [host.strip() for host in hosts.split(',') if host.strip()]
        for host in hostsList:
            # Dynamically retrieve the endpoint URL from the adapter instance
            metrics_url = f"{constants.HOST_PROTOCOL}://{host}/{constants.HOST_URL}"
            # Add the endpoint URL to the EndpointResult
            result.with_endpoint(metrics_url)
        
        logger.debug(f"Returning endpoints: {result.get_json()}")
        return result


# Main entry point of the adapter
def main(argv: List[str]) -> None:
    logging.setup_logging("adapter.log")
    # Start a new log file by calling 'rotate'. By default, the last five calls will be
    # retained. If the logs are not manually rotated, the 'setup_logging' call should be
    # invoked with the 'max_size' parameter set to a reasonable value, e.g.,
    # 10_489_760 (10MB).
    logging.rotate()

    logger.info(f"Running adapter code with arguments: {argv}")
    if len(argv) != 3:
        # `inputfile` and `outputfile` are always automatically appended to the
        # argument list by the server
        logger.error("Arguments must be <method> <inputfile> <ouputfile>")
        exit(1)

    method = argv[0]
    try:
        if method == "test":
            test(AdapterInstance.from_input()).send_results()
        elif method == "endpoint_urls":
            get_endpoints(AdapterInstance.from_input()).send_results()
        elif method == "collect":
            collect(AdapterInstance.from_input()).send_results()
        elif method == "adapter_definition":
            result = get_adapter_definition()
            if type(result) is AdapterDefinition:
                result.send_results()
            else:
                logger.info(
                    "get_adapter_definition method did not return an AdapterDefinition"
                )
                exit(1)
        else:
            logger.error(f"Command {method} not found")
            exit(1)
    finally:
        logger.info(Timer.graph())


if __name__ == "__main__":
    main(sys.argv[1:])
